from __future__ import annotations

__all__ = ["markup"]

import html
import string
from typing import Any, Iterator, Sequence, Optional
import xml.etree.ElementTree as ET


from .ansi import Ansi, AnsiNull

class XMLEscapeFormatter(string.Formatter):
    """
    If conversion specifier 'e' (escape) is
    given or omitted, escape xml contents of
    the value. If 'r' (raw) is given, insert
    without escaping.
    """
    
    @staticmethod
    def convert_field(value, conversion):
        if conversion in ["e", None]:
            return html.escape(value)
        
        elif conversion == "r":
            return value
        
        raise ValueError(f"Unknown conversion specifier {conversion}")

def get_ansi_from_tag(name: str, style: dict[str, Ansi] = {}) -> Ansi:
    if name in style:
        return style[name]
    
    raise ValueError(f"invalid tag {name!r}")

def style_element(
    element: ET.Element,
    style: dict[str, Ansi] = {},
    _previous_ansi: Ansi = AnsiNull(),
) -> str:
    """
    Yields
    ------
    Text snippets.
    """
    for child in element:
        tag = child.tag
        ansi = get_ansi_from_tag(tag, style = style)
        
        yield ansi.enable_str() + (child.text or "")
        
        yield from style_element(
            child,
            style = style,
            _previous_ansi = ansi
        )
        
        yield (
            ansi.disable_str() +
            _previous_ansi.enable_str() +
            (child.tail or "")
        )
    
    #? yield _previous_ansi.disable_str()

def markup(
    string: str,
    args: Sequence[Any] = [],
    kwargs: dict[str, Any] = {},
    style: Optional[dict[str, Ansi]] = None
) -> str:
    """
    Use XML to style a string. In order to
    format the string, provide the variables
    directly to this function. All variables
    will be escaped automatically unless ``!r``
    is added at the end. You can also use ``!e``
    to explicitly specify that the variable
    has to be escaped.
    
    Parameters
    ----------
    string
        The markup text.
    
    args
        Variables to insert positional.
    
    kwargs
        Variables to insert by keyword.
    
    style
        Mapping of keys and the style to use.
        You may want to use
        :function:`utils.retrieve_ansi(locals())`.
    
    Returns
    -------
    Styled text.
    """
    text = XMLEscapeFormatter().format(string, *args, **kwargs)
    
    root = ET.fromstring(f"<root>{text}</root>")
    text = root.text or ""
    text += "".join(style_element(root, style = style))
    
    return html.unescape(text)



from .ansi import Ansi, get_ansi_string

class Style(Ansi):
    def __init__(self, enable, disable):
        self._ansi = [enable]
        self._off = [disable]

BOLD = Style(1, 22)          # b
DIM = Style(2, 22)           # d
ITALIC = Style(3, 23)        # i
UNDERLINE = Style(4, 24)     # u
BLINK = Style(5, 25)         # f (flash)
INVERSE = Style(7, 27)       # r (reverse)
INVISIBLE = Style(8, 28)     # h (hidden)
STRIKETHROUGH = Style(9, 29) # s



from enum import auto, IntEnum
import os
import sys

class Terminal(IntEnum):
    NOCOLOR = auto()
    BIT3 = auto()
    BIT4 = auto()
    BIT8 = auto()
    BIT24 = auto()
    
    @classmethod
    def get_term(cls):
        if not (sys.stdout and sys.stdout.isatty()):
            raise RuntimeError("standard output is not a valid terminal")
        
        ac = os.getenv("ADORABLE_COLOR", None)
        if ac == "nocolor":
            return cls.NOCOLOR
        
        if ac == "3bit":
            return cls.BIT3
        
        if ac == "4bit":
            return cls.BIT4
        
        if ac == "8bit":
            return cls.BIT8
        
        if ac == "24bit":
            return cls.BIT24
        
        if int(os.getenv("NO_COLOR", "0")):
            return cls.NOCOLOR
        
        if os.getenv("COLORTERM", "0") in ["truecolor", "24bit"]:
            return cls.BIT24
        
        t = os.getenv("TERM", "").removeprefix("xterm-")
        if t == "256":
            return cls.BIT8
        
        return cls.NOCOLOR


from collections import namedtuple
from collections.abc import Mapping
from functools import wraps
from math import sqrt
from typing import Any, Callable, Sequence, TypeVar, Union
import warnings

from .ansi import Ansi

T = TypeVar("T")

RGB = namedtuple("RGB", "r g b")

T_RGB = Union[tuple[int, int, int], RGB]
HEX = Union[str, int]

def retrieve_ansi(style: Mapping[str, Any]) -> dict[str, Ansi]:
    return {k: v for k, v in style.items() if isinstance(v, Ansi)}

def get_closest_color(color: RGB, colors: Sequence[tuple[RGB, Any]]) -> int:
    diffs = {}
    for value, rgb in colors:
        if color == rgb:
            return value
        
        diff = sqrt(
            abs(color.r - rgb.r) ** 2 +
            abs(color.g - rgb.g) ** 2 +
            abs(color.b - rgb.b) ** 2
        )
        
        diffs[diff] = value
    
    return diffs[min(diffs.keys())]




from __future__ import annotations

from abc import ABC
from sys import stdout
from typing import Any, Optional, Sequence


class Ansi(ABC):
    def __str__(self):
        return self.enable_str()
    
    def __call__(self, *args, **kwargs):
        return paint(*args, style = self, **kwargs)
    
    def enable_str(self) -> str:
        return get_ansi_string(*self._ansi)
    
    def enable(self, file = stdout) -> None:
        file.write(self.enable_str())
    
    def disable_str(self) -> str:
        return get_ansi_string(*self._off)
    
    def disable(self, file = stdout) -> None:
        file.write(self.disable_str())

class AnsiNull(Ansi):
    def __init__(self):
        self._ansi = []
        self._off = []

def get_ansi_string(*args):
    return f"\x1b[{';'.join(map(str, args))}m"

def paint(*args: Any, style: Optional[Ansi] = None, sep: str = " ") -> str:
    content = sep.join(map(str, args))
    
    if style is None:
        return content
    
    elif not isinstance(style, Ansi):
        raise TypeError(f"expected `None`, `Ansi` for argument `style`, got {style.__class__.__name__}")
    
    return f"{style}{content}{style.disable_str()}"

def printc(*args: Any, **kwargs: Any) -> None:
    paint_kwargs = {}
    
    for key in ["style", "sep"]:
        if key in kwargs:
            paint_kwargs[key] = kwargs.pop(key)
    
    print(paint(*args,
        **paint_kwargs
    ), **kwargs)
